# README for utils/config directory

**This information is of value to developers only.**

## Basics:

Scripts in this directory are run to generate content in `upcxx_config.hpp`.
To be run, they must be named in the `UPCXX_CONFIG_SCRIPTS` variable in
`bld/config.mak`.

The interface is simple: inputs via environment (below) and
output on `stdout`.  Everything on `stdout` is sent to the config header.

All `stderr` is provided to the user. Exiting non-zero stops the build.

Scripts are run after GASNet has been built.  This should allow a very wide
variety of tests to be constructed.  However, be aware that it is not (in
general) possible to execute test codes when cross-compiling.

Scripts should assume multiple instances will be running concurrently for
different builds of libupcxx.a, with distinct working directories.  Other than
sepration of concurrent runs, scripts should not assume any specific working
directory and should remove any temporary files it creates.  The prefix
`conftest` is preferred for any temporary files (source, object, etc.) and it
is permissible to overwrite any pre-exiting files with that prefix (perhaps
from other probes that failed to cleanup).

## Environment:

While GNU Make may "leak" many of its variables into the environment, scripts
should only depend on ones documented here.

The following environment variables carry information about the UPC\+\+ library
instance to be built using the resulting `upcxx_config.hpp`:

* `UPCXX_TOPSRC`: top of the source directory hierarchy
* `UPCXX_TOPBLD`: top of the build directory hierarchy
* `UPCXX_ASSERT`: either `0` or `1`
* `UPCXX_OPTLEV`: either `0` or `3`
* `UPCXX_DBGSYM`: either `0` or `1`
* `UPCXX_NETWORK`: the conduit (lowercase)
* `UPCXX_CODEMODE`: either `opt` or `debug`
* `UPCXX_THREADMODE`: either `seq` or `par`
* `UPCXX_CROSS`: the cross-configure platform name, or empty

Additionally, `UPCXX_GMAKE` is defined to the GNU Make discovered by
`configure` and can be used to extract additional configure outputs.  The
following will set the shell variable `UPCXX_CUDA_NVCC` to the corresponding
value in the top-level `Makefile` generated by `configure`:

```bash
eval $($UPCXX_GMAKE -C "$UPCXX_TOPBLD" echovar VARNAME=UPCXX_CUDA_NVCC)
```

The following each carry the corresponding values from the makefile fragment:

* `GASNET_CC`
* `GASNET_CFLAGS`
* `GASNET_OPT_CFLAGS`
* `GASNET_MISC_CFLAGS`
* `GASNET_CXX`
* `GASNET_CXXFLAGS`
* `GASNET_OPT_CXXFLAGS`
* `GASNET_MISC_CXXFLAGS`
* `GASNET_CPPFLAGS`
* `GASNET_MISC_CPPFLAGS`
* `GASNET_CXXCPPFLAGS`
* `GASNET_MISC_CXXCPPFLAGS`
* `GASNET_DEFINES`
* `GASNET_INCLUDES`
* `GASNET_LD`
* `GASNET_LDFLAGS`
* `GASNET_LIBS`
